__version__ = "2023.1.0"
__nabu_modules__ = [
    "app",
    "cuda",
    "estimation",
    "io",
    "misc",
    "opencl",
    "pipeline",
    "preproc",
    "reconstruction",
    "resources",
    "thirdparty",
]
version = __version__
